#ifndef DS_BASICISTREAM__HPP
#define DS_BASICISTREAM__HPP

#include <ds_basicios.hpp>
#include <string>

namespace DS_NAMESPACE
{
  template <TYPENAME charT>
  class basic_istream : public virtual basic_ios<charT>
  {
  protected:
    size_t count;
    bool lastreadvalid;
    charT lastread;

    int getImp()
    {
      if(fail())
	return BIOSERR;
      if(lastreadvalid)
      {
	lastreadvalid=false;
	count++;
	return lastread;
      }
      int i=sbget();
      if(i==BIOSERR)
      {
	setstate(eofbit);
	return BIOSERR;
      }
      lastread=i;
      count++;
      return i;
    }

    void skipwhitespace()
    {
      bool flag=true;
      do
      {
	int c=getImp();
	if(!whitespace(c))
	{
	  unget();
	  flag=false;
	}
      } while(flag && good());
    }

  public:
    basic_istream<charT>(streambuf *s=0) :
      basic_ios<charT>(s),
      lastreadvalid(false)
    {
      ds_debughook("basic_istream()");
    }

    ~basic_istream<charT>()
    {
      ds_debughook("~basic_istream()");
    }

    streampos tellg() const
    {
      streampos p=0;
      if(sb_)
	p=sb_->pos();
      return p;
    }
    basic_istream<charT>& seekg(const streamoff offs)
    {
      if(sb_)
	sb_->seek(offs, ios::beg);
      return *this;
    }
    basic_istream<charT>& seekg(const streamoff offs, const ios::seek_dir flag)
    {
      if(sb_)
	sb_->seek(offs, flag);
      return *this;
    }

    int sync() const
    {
      if(sb_)
	return sb_->flush();
      return -1;
    }

    int get()
    {
      count=0;
      return getImp();
    }
    basic_istream<charT>& get(charT& c)
    {
      count=0;
      c=getImp();
      return *this;
    }
    basic_istream<charT>& get(charT* zf, size_t anz, charT ende='\n')
    {
      count=0;
      while(anz-- && good())
      {
	int i=getImp();
	if(i==BIOSERR || i==ende)
	  break;
	*zf++=i;
      }
      *zf=0;
      return *this;
    }
    basic_istream<charT>& getline(charT *zf, size_t anz, charT ende='\n')
    {
      count=0;
      while(anz-- && good())
      {
	int i=getImp();
	if(i==BIOSERR)
	  break;
	*zf++=i;
	if(i==ende)
	  break;
      }
      *zf=0;
      return *this;
    }
    basic_istream<charT>& read(charT *zf, size_t anz)
    {
      count=0;
      while(anz-- && good())
      {
	int i=getImp();
	if(i==BIOSERR)
	  break;
	*zf++=i;
      }
      if(rdstate()&eofbit)
	setstate(failbit);
      return *this;
    }
    basic_istream<charT>& readsome(charT *zf, size_t anz)
    {
      count=0;
      while(anz-- && good())
      {
	int i=getImp();
	if(i==BIOSERR)
	  break;
	*zf++=i;
      }
      return *this;
    }
    size_t gcount() const { return count; }
    basic_istream<charT>& ignore(int anz=1, int ende=EOF)
    {
      while(anz-- && good())
	if(getImp()==BIOSERR)
	  break;
      return *this;
    }
    int peek()
    {
      int i=getImp();
      if(i==BIOSERR)
	return BIOSERR;
      lastreadvalid=true;
      return i;
    }
    basic_istream<charT>& unget()
    {
      if(lastreadvalid)
	setstate(badbit);
      lastreadvalid=true;
      return *this;
    }
    basic_istream<charT>& putback(const charT c)
    {
      const bool thesame=(lastread==c);
      if(thesame)
      {
	if(lastreadvalid)
	  setstate(badbit);
	lastreadvalid=true;
      }
      return *this;
    }

    basic_istream<charT>& get(streambuf& s, const charT delim='\n')
    {
      for(;;)
	{
	  const int r=this->getImp();
	  if(r==BIOSERR)
	    break;
	  if(r==delim)
	    {
	      unget();
	      break;
	    }
	  const int rr=s.put(r);
	  if(rr==streambuf::SBEOF)
	    {
	      unget();
	      break;
	    }
	}

      return *this;
    }

    basic_istream<charT>& operator>> (bool& b)
    {
      if(fmtflags_&boolalpha)
      {
	std::basic_string<charT> s;
	operator>> (s);
	basic_ios_traits<charT> t;
	if(s==t.truestring)
	  b=true;
	else if(s==t.falsestring)
	  b=false;
	else
	  setstate(badbit);
      }
      else
      {
	count=0;
	skipwhitespace();
	// get the value
	int c=getImp();
	if(c==BIOSERR)
	  return *this;
	if(c=='0')
	  b=false;
	else
	  b=true;
	// skip any more thing until we reach whitespace again
	do
	{
	  c=getImp();
	} while((!whitespace(c)) && good());
	unget();
      }
      return *this;
    }

    basic_istream<charT>& get(std::basic_string<charT>& s)
    {
      // todo
      return *this;
    }

    basic_istream<charT>& operator>> (unsigned int& num)
    {
      count=0;
      num=0;
      skipwhitespace();
      if(fail())
	return *this;

      int multiplier=10;
      int c=getImp();
      if(c==BIOSERR)
	return *this;
      if(c=='+')
	c=getImp(); // '+' consumed
      if(c==BIOSERR)
	return *this;
      // leading zero indicates either octal or hexadecimal
      if(c=='0')
      {
	// first set octal
	multiplier=8;
	c=getImp();
	if(c==BIOSERR)
	  return *this;
	// check for hex
	if(c=='x' || c=='X')
	  multiplier=16;
	else
	  // no, no hex :-( you octal suckers
	  unget();
      }
      else
	// ok, we take the beloved 10
	unget();

      // read number
      do {
	c=getImp();
	if(c==BIOSERR)
	  return *this;
	if(whitespace(c))
	{
	  unget();
	  break;
	}
	// check valid characters along with base
	if(c>='0' && c<='7')
	  c-='0';
	else if(c>='8' && c<='9' && multiplier<=10)
	  c-='0';
	else if(c>='a' && c<='f' && multiplier<=16)
	  c-='a'+10;
	else if(c>='A' && c<='F' && multiplier<=16)
	  c-='A'+10;
	else
	  break; // invalid character for this mode

	num*=multiplier	;
	num+=c;
      } while(good());

      // read until we reach whitespace again
      do {
	c=getImp();
      } while((!whitespace(c)) && good());

      return *this;
    }
    basic_istream<charT>& operator>> (int& num)
    {
      count=0;
      num=0;
      skipwhitespace();
      if(fail())
	return *this;
      
      // check for '-'
      bool neg=false;
      int c=getImp();
      if(c==BIOSERR)
	return *this;
      if(c=='-')
	neg=true;
      else
	unget();

      unsigned int n;
      operator>> (n);

      num=static_cast<int>(n);
      if(neg)
	num=-num;

      return *this;
    }
    basic_istream<charT>& operator>> (charT *str)
    {
      count=0;
      skipwhitespace();

      int c;
      do {
	c=getImp();
	if(c==BIOSERR)
	  return *this;
	if(whitespace(c))
	  break;
	*str++=c;
      } while(good());

      return *this;
    }
    basic_istream<charT>& operator>> (std::basic_string<charT>& str)
    {
      str.erase();
      count=0;
      skipwhitespace();

      int c;
      do {
	c=getImp();
	if(c==BIOSERR)
	  return *this;
	if(whitespace(c))
	  break;
	str+=c;
      } while(good());

      return *this;
    }
    basic_istream<charT>& operator>> (void* &v)
    {
      unsigned int i;
      operator>>(i);
      v=reinterpret_cast<void*>(i);
      return *this;
    }

    // iomanip overloads

    basic_istream<charT>& operator>> (resetiosflags_t mask)
    {
      this->unsetf(mask.flags);
      return *this;
    }

    basic_istream<charT>& operator>> (setiosflags_t mask)
    {
      this->setf(mask.flags);
      return *this;
    }

    basic_istream<charT>& operator>> (setbase_t base)
    {
      switch(base.base)
	{
	case 8:  this->setf(oct, basefield); break;
	case 10: this->setf(dec, basefield); break;
	case 16: this->setf(hex, basefield); break;
	default: this->setf(0,   basefield); break;
	}
      return *this;
    }
    
    basic_istream<charT>& operator>> (setfill_t f)
    {
      this->fill(f.fill);
      return *this;
    }
    
    basic_istream<charT>& operator>> (setprecision_t p)
    {
      this->precision(p.prec);
      return *this;
    }
    
    basic_istream<charT>& operator>> (setw_t w)
    {
      this->width(w.width);
      return *this;
    }
    
  };

  template <TYPENAME charT>
  basic_istream<charT>& getline(basic_istream<charT>& is, std::basic_string<charT>& str, const charT delim='\n')
  {
    do
      {
	int c=is.get();
	if(c==delim || is.fail())
	  break;
	str+=static_cast<charT>(c);
      }
    while(true);
    return is;
  }

#if 0
  template <TYPENAME charT>
  basic_istream<charT>& operator>> (basic_istream<charT>& is, basic_string<charT>& s)
  {
    return is.get(s);
  }
#endif

  typedef basic_istream<char> istream;
  typedef basic_istream<wchar_t> wistream;
}

#endif
