#ifndef DS_BASICOSTREAM__HPP
#define DS_BASICOSTREAM__HPP

#include <ds_basicios.hpp>
#include <string>

namespace DS_NAMESPACE
{
  template <TYPENAME charT>
  class basic_ostream : public virtual basic_ios<charT>
  {
  protected:
    void putImp(const charT c)
    {
      if(good())
      {
	int r=this->sbput(c);
	if(r==BIOSERR)
	  this->setstate(failbit);
	else
	  this->clear();
      }
    }

    void putNum(const std::basic_string<charT>& s)
    {
      typename std::basic_string<charT>::const_iterator i=s.end();
      for(--i; i>=s.begin() && good(); --i)
	this->putImp(*i);
    }

    void putFill(const size_t ssize)
    {
      if(width_>static_cast<streamsize>(ssize))
	{
	  width_-=ssize;
	  while(width_--)
	    this->putImp(fill_);
	}
      width_=0;
    }

    void putStr(const std::basic_string<charT>& s)
    {
      for(typename std::basic_string<charT>::const_iterator i=s.begin(); i!=s.end(); ++i)
	this->putImp(*i);
    }

    void putUnsigned(unsigned int num, charT prefix)
    {
      const ios::fmtflags sav=fmtflags_;

      // fix fmtflags_
      const unsigned f=fmtflags_&basefield;
      if(f==0 || f==(oct|dec) || f>hex)
	this->setf(dec, basefield);

      if((fmtflags_&dec) && (fmtflags_&showpos))
	prefix='+';

      this->fixAdjustFlags();

      typedef std::basic_string<charT> s_t;
      s_t s;
      s_t prefixStr;
      if(prefix)
	prefixStr+=prefix;

      // shortcut for 0
      if(num==0)
	{
	  s="0";
	}
      else
	{
	  // get character table
	  basic_ios_traits<charT> traits;
	  charT *tab=(fmtflags_&uppercase)?traits.uppernumber:traits.lowernumber;

	  // set divisor
	  const int divisor=(fmtflags_&hex)?16:( (fmtflags_&oct)?8:10 );

	  // and now the number
	  do {
	    s+=tab[num%divisor];
	    num/=divisor;
	  } while(num);

	  // print base
	  if(fmtflags_&showbase)
	    {
	      if(divisor==8)
		s+='0';
	      else if(divisor==16)
		{
		  prefixStr+='0';
		  prefixStr+=tab[16]; // x
		}
	    }

	}

      if(fmtflags_&right)
	{
	  this->putFill(s.size()+prefixStr.size());
	  this->putStr(prefixStr);
	  this->putNum(s);
	}
      else if(fmtflags_&left)
	{
	  this->putStr(prefixStr);
	  this->putNum(s);
	  this->putFill(s.size()+prefixStr.size());
	}
      else if(fmtflags_&internal)
	{
	  this->putStr(prefixStr);
	  this->putFill(s.size()+prefixStr.size());
	  this->putNum(s);
	}

      width_=0;
      fmtflags_=sav;
    }

    void fixAdjustFlags()
    {
      const unsigned f=fmtflags_&adjustfield;
      if(f==0 || f==(left|right) || f>internal)
	this->setf(right, adjustfield);
    }

  public:
    basic_ostream<charT>(streambuf *s=0) :
      basic_ios<charT>(s)
    {
      ds_debughook("basic_ostream()");
    }

    ~basic_ostream<charT>()
    {
      ds_debughook("~basic_ostream()");
    }

    streampos tellp() const
    {
      streampos p=0;
      if(sb_)
	p=sb_->pos();
      return p;
    }

    basic_ostream<charT>& seekp(const streampos p)
    {
      if(sb_)
	sb_->seek(p, ios::beg);
      return *this;
    }

    basic_ostream<charT>& seekp(const streamoff offs, const ios::seek_dir flag)
    {
      if(sb_)
	sb_->seek(offs, flag);
      return *this;
    }

    basic_ostream<charT>& put(const charT c)
    {
      if(width_==0)
	this->putImp(c);
      else
	{
	  const ios::fmtflags sav=fmtflags_;
	  this->fixAdjustFlags();
	  if(fmtflags_&left)
	    this->putImp(c);
	  while(--width_)
	    this->putImp(fill_);
	  if((fmtflags_&right) || (fmtflags_&internal))
	    this->putImp(c);
	  fmtflags_=sav;
	}
      return *this;
    }
    basic_ostream<charT>& write(const charT* zf, size_t anz)
    {
      return this->operator<< ( std::basic_string<charT> (zf, zf+anz) );
    }

    basic_ostream<charT>& flush()
    {
      if(sb_)
	sb_->flush();
      else
	this->setstate(badbit);
      return *this;
    }

    basic_ostream<charT>& operator<< (const charT c)
    {
      return this->put(c);
    }

    basic_ostream<charT>& operator<< (const bool b)
    {
      if(fmtflags_&boolalpha)
      {
	basic_ios_traits<charT> t;
	return this->operator<< (b?t.truestring:t.falsestring);
      }
      return this->put(b?'1':'0');
    }

    basic_ostream<charT>& operator<< (unsigned int num)
    {
      this->putUnsigned(num, 0);
      return *this;
    }

    basic_ostream<charT>& operator<< (int num)
    {
      const ios::fmtflags fmtsav=fmtflags_&showpos; // save state of showpos

      // oct and hex don't make negative output, they interpret themselfs as unsigned
      if(fmtflags_&hex || fmtflags_&oct)
	{
	  this->operator<< (*reinterpret_cast<unsigned int*>(&num));
	}
      // dec output
      else
	{
	  charT prefix=0;
	  if(num<0)
	    {
	      prefix='-';
	      num=-num;
	      fmtflags_&=~showpos; // clear showpos, we don't need an additional '+'
	    }
	  this->putUnsigned(static_cast<unsigned int>(num), prefix);
	}
      fmtflags_|=fmtsav; // restore showpos
      return *this;
    }
    basic_ostream<charT>& operator<< (const charT *str)
    {
      return this->operator<< (std::basic_string<charT>(str));
    }
    basic_ostream<charT>& operator<< (const std::basic_string<charT>& str)
    {
      const ios::fmtflags sav=fmtflags_;
      this->fixAdjustFlags();

      if((fmtflags_&right) || (fmtflags_&internal))
	this->putFill(str.size());

      this->putStr(str);      

      if(fmtflags_&left)
	this->putFill(str.size());

      fmtflags_=sav;
      width_=0;
      return *this;
    }
    basic_ostream<charT>& operator<< (const void *v)
    {
      const ios::fmtflags fmtsav=fmtflags_;
      fmtflags_|=uppercase;
      fmtflags_|=showbase;
      fmtflags_|=hex;
      this->operator<< (reinterpret_cast<unsigned int>(v));
      fmtflags_=fmtsav;
      return *this;
    }

    basic_ostream<charT>& operator<< (streambuf *sb)
    {
      if(sb)
	while(this->good())
	  {
	    const int c=sb->get();
	    if(c==streambuf::SBEOF)
	      break;
	    this->putImp(static_cast<charT>(c));
	  }
      width_=0;
      return *this;
    }

    /// manipulator overload
    basic_ostream<charT>& operator<< ( basic_ostream<charT>& (*func) (basic_ostream<charT>&) )
    {
      return func(*this);
    }

    basic_ostream<charT>& operator<< ( ios_base& (*func) (ios_base&) )
    {
      func(*this);
      return *this;
    }

    basic_ostream<charT>& operator<< ( basic_ios<charT>& (*func) (basic_ios<charT>&) )
    {
      func(*this);
      return *this;
    }

    // iomanip overloads

    basic_ostream<charT>& operator<< (resetiosflags_t mask)
    {
      this->unsetf(mask.flags);
      return *this;
    }

    basic_ostream<charT>& operator<< (setiosflags_t mask)
    {
      this->setf(mask.flags);
      return *this;
    }

    basic_ostream<charT>& operator<< (setbase_t base)
    {
      switch(base.base)
	{
	case 8:  this->setf(oct, basefield); break;
	case 10: this->setf(dec, basefield); break;
	case 16: this->setf(hex, basefield); break;
	default: this->setf(0,   basefield); break;
	}
      return *this;
    }
    
    basic_ostream<charT>& operator<< (setfill_t f)
    {
      this->fill(f.fill);
      return *this;
    }
    
    basic_ostream<charT>& operator<< (setprecision_t p)
    {
      this->precision(p.prec);
      return *this;
    }
    
    basic_ostream<charT>& operator<< (setw_t w)
    {
      this->width(w.width);
      return *this;
    }
    

  };

  typedef basic_ostream<char> ostream;
  typedef basic_ostream<wchar_t> wostream;

  // manipulators

  template <TYPENAME charT>
  inline basic_ostream<charT>& endl(basic_ostream<charT>& os)
  {
    return os << '\n' << flush;
  }

  template <TYPENAME charT>
  inline basic_ostream<charT>& ends(basic_ostream<charT>& os)
  {
    return os << '\0';
  }

  template <TYPENAME charT>
  inline basic_ostream<charT>& flush(basic_ostream<charT>& os)
  {
    return os.flush();
  }
#if 0
  template <TYPENAME charT>
  basic_ostream<charT>& operator<< (std::basic_ostream<charT>& os, const std::basic_string<charT>& s)
  {
    return os << s.c_str();
  }
#endif
}

#endif
