#ifndef IOSBASE__HPP
#define IOSBASE__HPP

#include <ds_config.hpp>

namespace DS_NAMESPACE
{

  typedef long streampos;
  typedef long streamoff;
  typedef int streamsize;

  class ios_base
  {
  public:
    enum
    {
      boolalpha = 0x01,
      showbase =  0x02,
      showpoint = 0x04,
      showpos =   0x08,
      skipws =    0x10,
      unitbuf =   0x20,
      uppercase = 0x40,

      left =	  0x100,
      right =	  0x200,
      internal =  0x400,
      adjustfield=left|right|internal,

      oct =	  0x1000,
      dec =	  0x2000,
      hex =	  0x4000,
      basefield = oct|dec|hex,

      fixed =	  0x10000,
      scientific= 0x20000,
      floatfield= fixed|scientific
    };

    enum openmode
    {
      in=1,
      out=2,
      app=4,
      ate=8,
      trunc=16,
      binary=32,
      nocreate=64,
      noreplace=128,
    };

    enum seek_dir { beg, cur, end };

    enum iostate
    {
      goodbit = 1,
      eofbit =  2,
      failbit = 4,
      badbit =  8
    };

    iostate state_;

    typedef unsigned int fmtflags;

  protected:
    bool whitespace(int c)
    {
      return (c==' ')||(c=='\t')||(c=='\n')||(c=='\r');
    }

    streamsize width_;
    int fill_;
    streamsize precision_;

    fmtflags fmtflags_;

  public:
    ios_base() :
      state_(goodbit),
      width_(0),
      fill_(' '),
      precision_(6),
      fmtflags_(skipws|right|dec)
    {
    };

    int good() { return this->rdstate()&goodbit; }
    int eof() { return this->rdstate()&eofbit; }
    int fail() { return this->rdstate()&(failbit|badbit); }
    int bad() { return this->rdstate()&badbit; }

    int operator! () { return this->fail(); }

    void clear(iostate state = goodbit) { this->setstate(state); }
    void setstate(iostate state) { state_=state; }
    iostate rdstate() const { return state_; }

    operator void* () { return this->good()?this:0; }

    int fill(int f)
    { 
      const int oldfill_=fill_;
      fill_=f;
      return oldfill_;
    }

    fmtflags flags() const { return fmtflags_; }
    fmtflags flags(fmtflags n)
    {
      const fmtflags of=fmtflags_;
      fmtflags_=n;
      return of;
    }

    fmtflags setf(fmtflags n)
    { 
      const fmtflags of=fmtflags_;
      fmtflags_|=n;
      return of;
    }
    fmtflags setf(fmtflags n, fmtflags mask)
    {
      const fmtflags of=fmtflags_;
      fmtflags_= (mask & n) | (of & (~mask));
      return of;
    }
    void unsetf(fmtflags mask)
    {
      fmtflags_ &= ~mask;
    }

    streamsize precision() const { return precision_; }
    streamsize precision(streamsize np)
    {
      const streamsize op=precision_;
      precision_=np;
      return op;
    }

    streamsize width() const { return width_; }
    streamsize width(streamsize nw)
    {
      const streamsize ow=width_;
      width_=nw;
      return ow;
    }

    static void sync_with_stdio(bool sync=true);

    class Init
    {
    public:
      Init();
      static void Realinit();
    };
  };


  // manipulators
  inline ios_base& boolalpha(ios_base& i)   { i.setf(ios_base::boolalpha); return i; }
  inline ios_base& noboolalpha(ios_base& i) { i.unsetf(ios_base::boolalpha); return i; }

  inline ios_base& showbase(ios_base& i)   { i.setf(ios_base::showbase); return i; }
  inline ios_base& noshowbase(ios_base& i) { i.unsetf(ios_base::showbase); return i; }

  inline ios_base& showpoint(ios_base& i)   { i.setf  (ios_base::showpoint); return i; }
  inline ios_base& noshowpoint(ios_base& i) { i.unsetf(ios_base::showpoint); return i; }

  inline ios_base& showpos(ios_base& i)   { i.setf  (ios_base::showpos); return i; }
  inline ios_base& noshowpos(ios_base& i) { i.unsetf(ios_base::showpos); return i; }

  inline ios_base& skipws(ios_base& i)   { i.setf  (ios_base::skipws); return i; }
  inline ios_base& noskipws(ios_base& i) { i.unsetf(ios_base::skipws); return i; }

  inline ios_base& unitbuf(ios_base& i)   { i.setf  (ios_base::unitbuf); return i; }
  inline ios_base& nounitbuf(ios_base& i) { i.unsetf(ios_base::unitbuf); return i; }

  inline ios_base& uppercase(ios_base& i)   { i.setf  (ios_base::uppercase); return i; }
  inline ios_base& nouppercase(ios_base& i) { i.unsetf(ios_base::uppercase); return i; }

  inline ios_base& left(ios_base& i)     { i.setf(ios_base::left, ios_base::adjustfield); return i; }
  inline ios_base& right(ios_base& i)    { i.setf(ios_base::right, ios_base::adjustfield); return i; }
  inline ios_base& internal(ios_base& i) { i.setf(ios_base::internal, ios_base::adjustfield); return i; }

  inline ios_base& oct(ios_base& i) { i.setf(ios_base::oct, ios_base::basefield); return i; }
  inline ios_base& dec(ios_base& i) { i.setf(ios_base::dec, ios_base::basefield); return i; }
  inline ios_base& hex(ios_base& i) { i.setf(ios_base::hex, ios_base::basefield); return i; }

  inline ios_base& fixed(ios_base& i)      { i.setf(ios_base::fixed, ios_base::floatfield); return i; }
  inline ios_base& scientific(ios_base& i) { i.setf(ios_base::scientific, ios_base::floatfield); return i; }

}

#include <ds_iomaniptypes.hpp>

#endif
