// Copyright (c) 2003 Dirk Jagdmann <doj@cubic.org> and Fabian
// Giesen <ryg@gmx.net>
//
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any
// damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any
// purpose, including commercial applications, and to alter it and
// redistribute it freely, subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented; you must
// not claim that you wrote the original software. If you use this
// software in a product, an acknowledgment in the product
// documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such, and must
// not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source
// distribution.

#include <iostream>

static std::ostream& operator<< (std::ostream& os, const std::wstring& str)
{
  for(std::wstring::const_iterator i=str.begin(); i!=str.end(); ++i)
    {
      wchar_t c=*i;
      if(c<256)
      {
	unsigned char ch=static_cast<unsigned char>(c);
	os << *reinterpret_cast<char*>(&ch);
      }
      else
	os << '.';
    }
  return os;
}

#include "xml.hpp"

XML::XML(XML *par, const std::wstring &type) :
  parent(par),
  firstChild(0),
  lastChild(0),
  next(0),
  nodeType(type)
{ }

XML::~XML()
{
  if (next)
    {
      delete next;
      next=0;
    }

  if (firstChild)
    {
      delete firstChild;
      firstChild=0;
    }
}

XML* XML::addChild(const std::wstring &type)
{
  XML *node=new XML(this, type);

  if (lastChild)
    {
      lastChild->next=node;
      lastChild=node;
    }

  if (!firstChild)
    firstChild=lastChild=node;

  return node;
}


XML::Tokenizer::Tokenizer(std::istream &src) : 
  sourceStream(src), 
  curCh(-1), 
  lineCount(1), 
  tokenLineCount(0)
{
}

XML::Tokenizer::tokenType XML::Tokenizer::getToken()
{
  eatWhitespace();

  tokenLineCount=lineCount;

  if (sourceStream.eof())
    return tokenEOF;

  addInfo=L""; // clear additional info
  wchar_t ch=getChar();

  if (ch=='<')
    {
      bool end=false;

      while (!eof() && !end)
	{
	  ch=getChar();
	  if (ch=='>')
	    end=true;
	  else
	    addInfo+=ch;
	}

      end=false;
      const wchar_t firstChar=addInfo[0];
      if (firstChar=='/')
	{
	  addInfo.erase(0, 1);
	  end=true;
	}

      wchar_t lastChar=addInfo[addInfo.size()-1];
      if (lastChar=='/')
	{
	  addInfo.erase(addInfo.size()-1, 1);
	  return tokenStartEndTag;
	}

      if (firstChar=='?' && lastChar=='?')
	{
	  return tokenQuestionMark;
	}

      if (firstChar=='!')
	{
	  return tokenExclamationMark;
	}

      return end ? tokenEndTag : tokenStartTag;
    }
  else
    {
      addInfo=ch;

      // tokenize string

      bool tag=false;
      wchar_t ch=' ';
      
      while (!eof() && !tag)
	{
	  ch=getChar();
	  tag=(ch=='<');

	  if (!tag)
	    addInfo+=ch;
	}

      if (tag)
        unGetChar(ch);

      eliminateEscapes(addInfo);

      return tokenString;
    }
}

void XML::Tokenizer::eatWhitespace()
{
  bool white=true;

  while (!eof() && white)
    {
      const wchar_t ch=getChar();
      white=(ch==' ') || (ch=='\t') || (ch=='\r') || (ch=='\n');

      if (!white)
        unGetChar(ch);
    }
}

wchar_t XML::Tokenizer::getChar()
{
  wchar_t ch=' ';

  if (curCh==-1)
    ch=sourceStream.get();
  else
    {
      ch=curCh;
      curCh=-1;
    }

  if (ch=='\n')
    lineCount++;

  return ch;
}

void XML::Tokenizer::unGetChar(const wchar_t ch)
{
  assert(curCh == -1);

  if (ch=='\n')
    lineCount--;

  curCh=ch;
}

bool XML::Tokenizer::eof()
{
#pragma warning ( push )
#pragma warning ( disable : 4800 )
  return (curCh==-1) ? sourceStream.eof() : false;
#pragma warning ( pop )
}

void XML::Tokenizer::eliminateEscapes(std::wstring &str)
{
  size_t charInd=0;

  while (charInd<str.size())
    {
      if (str[charInd]=='&') // escape?
	{
	  int len=0;

	  while ((charInd+1+len)<str.size() && len<7 && str[charInd+len+1]!=';')
	    len++;

	  if (str[charInd+len+1]!=';')
	    std::clog << "XMLP: invalid escape in line " << lineCount << std::endl;
	  else // we found an escape!
	    {
	      std::wstring escape=str.substr(charInd+1, len);
	      wchar_t newCh=0;

	      if (escape==L"lt")
		newCh='<';
	      else if (escape==L"gt")
		newCh='>';
	      else if (escape==L"amp")
		newCh='&';
	      else if (escape==L"quot")
		newCh='\'';
	      else
	      {
		std::clog << "XMLP: unknown entity ";
		std::string s; std::unicode2latin(escape, s); std::clog << s;
		std::clog << " in line " << lineCount << std::endl;
	      }

	      if (newCh)
		{
		  str.erase(charInd+1, len+1);
		  str[charInd]=newCh;
		}
	    }
	}

      charInd++;
    }
}

std::wstring XML::Tokenizer::getAdditionalInfo() const
{
  return addInfo;
}

int XML::Tokenizer::getTokenLineCount() const
{
  return tokenLineCount;
}

XML* XML::parse(std::istream &src)
{
  if(src.fail())
    return 0;

  Tokenizer tokenizer(src);

  std::wstring rstr=L"/";
  XML *root=new XML(0, rstr); // create root node
  XML *currentNode=root;

  Tokenizer::tokenType token=Tokenizer::tokenNone;
  while (token != Tokenizer::tokenEOF)
    {
      token=tokenizer.getToken();

      switch (token)
	{
	case Tokenizer::tokenStartTag:
	  currentNode=currentNode->addChild(tokenizer.getAdditionalInfo());
	  break;

	case Tokenizer::tokenEndTag:
	  {
	    XML *curParent=currentNode->parent;

	    if (curParent && tokenizer.getAdditionalInfo()==currentNode->nodeType.c_str())
	      currentNode=currentNode->parent;
	    else
	      std::clog << "XMLP: mismatched terminating tag in line " << tokenizer.getTokenLineCount() << std::endl;
	  }
	  break;

	case Tokenizer::tokenStartEndTag:
	  currentNode->addChild(tokenizer.getAdditionalInfo());
	  break;

	case Tokenizer::tokenString:
	  currentNode->nodeData+=tokenizer.getAdditionalInfo();
	  break;

	case Tokenizer::tokenQuestionMark:
	  currentNode->addQuestionMark(tokenizer.getAdditionalInfo());
	  break;

	case Tokenizer::tokenExclamationMark:
	  currentNode->addExclamationMark(tokenizer.getAdditionalInfo());
	  break;

	default:
	  break;
	}
    }
#if 0
  // strip of the fake root node
  currentNode=root->firstChild;

  root->parent=root->firstChild=root->lastChild=root->next=0;
  delete root;

  // strip of ? and ! nodes
  while(currentNode && (currentNode->nodeType[0]=='?' || currentNode->nodeType[0]=='!'))
    currentNode=currentNode->next;

  return currentNode;
#else
  return root;
#endif
}

static void printindent(std::ostream& os, const int indent)
{
  for(int i=0; i<indent; i++)
    os << " ";
}

static std::wstring escape(std::wstring s)
{
  size_t pos;
  for(pos=0; pos<s.size(); pos++)
    if(s[pos]=='&')
      {
	s.erase(pos, 1);
	s.insert(pos, L"&amp;");
      }
  while((pos=s.find('>')) != std::string::npos)
    {
      s.erase(pos, 1);
      s.insert(pos, L"&gt;");
    }
  while((pos=s.find('<')) != std::string::npos)
    {
      s.erase(pos, 1);
      s.insert(pos, L"&lt;");
    }
  return s;
}

void XML::print(std::ostream& os, const int indent)
{
  // the root node has nothing to print
  if(parent)
    {
      printindent(os, indent);
      os << "<";
      os << nodeType;
      os << ">";
      os << escape(nodeData);
    }
  for(v_t::const_iterator q=questionmark.begin(); q!=questionmark.end(); ++q)
    {
      printindent(os, indent);
      os << "<";
      os << *q;
      os << ">";
      os << std::endl;
    }
  for(v_t::const_iterator e=exclamationmark.begin(); e!=exclamationmark.end(); ++e)
    {
      printindent(os, indent);
      os << "<";
      os << *e;
      os << ">";
      os << std::endl;
    }
  if(firstChild)
    {
      os << std::endl;
      for(XML *child=firstChild; child; child=child->next)
	child->print(os, indent+2);
      printindent(os, indent);
    }
  if(parent)
  {
    os << "</";
    os << nodeType;
    os << ">";
    os << std::endl;
  }
}
